"""
This is another attempt at writing a lexer class for PyLR.

Upon matching text, it must execute a function which will cause it
to return a 2-tuple of type (tok, val) where token is an integer and
val is just any python object that will later be passed as an argument 
to the functions that the parser will call when it reduces. For Example

for the grammar

 E-> E + T
 E -> T
 T -> T * F
 T -> F
 F ->( E )
 F -> id

it is likely that the lexer should return the token value of id <tok> and
the integer value of id (string.atoi(id)).

In addition, the lexer must always return (eof, something else) when it's done
scanning to get the parser to continue to be called until parsing is done.
"""

import re, string

__version__ = "$Id: Lexer.py,v 0.2 1997/12/13 03:10:06 scott Exp scott $"

class PyLRSyntaxError(SyntaxError):
    pass

SKIPTOK = 0x01 # don't consider this a token that is to be considered a part of the grammar, like '\n'
MAPTOK = 0x02  # use the token associated with the pattern when it matches
EXECFN= 0x04   # execute the function associated with the pattern when it matches
USETXT = 0x08  # if you match a single character and want its ascii value to be the token

class Lexer:


    def __init__(self):
	self.tokenmap = {}
	self.prgmap = {}
	self.patlist = []
	self.lasttok = -1
	self.text = ""
	self.textindex = 0
	self.groupcount = 1

    def seteof(self):
	self.eof = max(self.tokenmap.values()) + 1

    def nexttok(self):
	self.lasttok = self.lasttok + 1
	return self.lasttok

    def settext(self, t):
	self.text = t
	self.textindex = 0

    def addpat(self, pat, tokname="", func=None, flags=MAPTOK|EXECFN):
	self.patlist.append(pat)
	ngroups = string.count(pat,  "(") - string.count(pat, "\(")
	tok = -2
	if not func:
	    flags = flags & ~EXECFN
	if not tokname:
	    flags = flags & ~MAPTOK
	if flags & MAPTOK:
	    self.tokenmap[tokname] = tok = self.nexttok()
	else:
	    tok = self.nexttok()
	self.prgmap[self.groupcount] = tok, flags, func
	self.groupcount = tgroups = self.groupcount + ngroups + 1

	
    def gettok(self, tok):
	return self.tokenmap[tok]


    def compile(self):
	self.seteof()
	for pi in range(len(self.patlist)):
	    self.patlist[pi] = '(' + self.patlist[pi] + ')'
	pat = string.join(self.patlist, "|")
	self.prg = re.compile(pat)
	    

    def scan(self):
	if self.textindex >= len(self.text):
	    return (self.eof, "")
	mo = self.prg.match(self.text, self.textindex)
	if not mo: 
	    raise PyLRSyntaxError, self.text[self.textindex:self.textindex + 10]
	self.textindex = self.textindex + len(mo.group(0))
	for group_index in self.prgmap.keys():
	    mtext = mo.group(group_index)
	    if mtext is not None:
		t, flags, fn = self.prgmap[group_index]
		self.match_group = group_index
		tmpres = mo.group(0)
		if flags & EXECFN:
		    tmpres = apply(fn, (mo,))
		if flags & USETXT:
		    t = ord(mo.group(0)[0])
		if flags & SKIPTOK:
		    return self.scan()
		else:
		    return (t, tmpres)

	    
    def interactivetest(self, verbose=0):
	if verbose: import time
	print "lexer interactive test outputs tuples of (token, val), -type 'q' to quit"
	toks = [""] * (self.lasttok + 1)
	for tn in self.tokenmap.keys():
	    toks[self.tokenmap[tn]] = tn
	while 1:
	    line = raw_input("> ")
	    if line == "q" or not line:
		return
	    self.settext(line)
	    start = time.time()
	    while 1:
		res = self.scan()
		if res[0] == self.eof:
		    break
		t, v = res
		t = toks[t]
		res = t, v
		print res
	    end = time.time()
	    if verbose: print "matched in %02f secs" % (end - start)

    def interact(self):
	"""just here as a shortcut to .interactivetest()"""
	self.interactivetest(1)

